/*==============================================================================
*NOME:   T1AMNPAL.C
*
*  Projeto: Questo A do Teste 1 de Mtodos Numricos Para lgebra Linear
*
*  Verso:              V 1.00
*  Autor :   Juliana Carpes Imperial        - 9714169-4
*
*Descrio do mdulo de implementao
*     Objetivos:
*        - Implementar uma funo para calcular a soluo de um sistema linear
*          por decomposio LU;
*        - Implementar um programa principal e funes auxiliares que, dado um
*          sistema, aplica a funo acima citadas e imprime resultados.
*=============================================================================*/

#include <stdio.h>

#define I_MAX_TAM  12    /* tamanho mximo da matriz */
#define SZ_NAO_UNI "Sistema nao tem solucao unica\n!!"

/* definio das funes */

/*==============================================================================
*Nome: iLU
*Descrio da Funo: funo para calcular a soluo de um sistema linear por
*                     decomposio LU.
*
*Parametros recebidos
*   A:  A de AX = B;
*   B:  B de AX = B;
*   iN: o tamanho do sistema.
*
*Parmetros retornados
*   X:  a soluo do sistema se ela for nica;
*   LU: as matrizes L e U de A = LU em uma s, para economizar espao dos zeros.
*
*Valor retornado
*   1: nica soluo
*   0: zero ou infinitas solues.
*=============================================================================*/

int iLU( double A[ ][ I_MAX_TAM ]  ,
         double B[ ]               ,
         double X[ ]               ,
         double LU[ ][ I_MAX_TAM ] ,
         int    iN                  ) ;


/*==============================================================================
*Nome: ImprimirL
*Descrio da Funo: funo para imprimir a matriz L de L\U.
*
*Parametros recebidos
*   LU: as matrizes L e U de A = LU em uma s, para economizar espao dos zeros;
*   iN: o tamannho da matriz L.
*=============================================================================*/

void ImprimirL( double LU[ ][ I_MAX_TAM ] ,
                int    iN                  ) ;


/*==============================================================================
*Nome: ImprimirU
*Descrio da Funo: funo para imprimir a matriz U de L\U.
*
*Parametros recebidos
*   LU: as matrizes L e U de A = LU em uma s, para economizar espao dos zeros;
*   iN: o tamannho da matriz U.
*=============================================================================*/

void ImprimirU( double LU[ ][ I_MAX_TAM ] ,
                int    iN                  ) ;


/*==============================================================================
*Nome: ImprimirVetor
*Descrio da Funo: funo para imprimir um vetor.
*
*Parametros recebidos
*   V:  o vetor;
*   iN: o tamannho do vetor.
*=============================================================================*/

void ImprimirVetor( double V[ ] ,
                    int    iN    ) ;


/* implementao das funes */

/*==============================================================================
*Nome: iLU
*=============================================================================*/

int iLU( double A[ ][ I_MAX_TAM ]  ,
         double B[ ]               ,
         double X[ ]               ,
         double LU[ ][ I_MAX_TAM ] ,
         int    iN                  ) {

   int    i                                , /* contador para linhas */
          j                                , /* contador para colunas */
          k                                , /* contador para somatrios */
          iProximo                     = 1 ; /* linha inicial para permutao */
   double Aux[ I_MAX_TAM ][ I_MAX_TAM ]    , /* matriz auxiliar */
          Bux[ I_MAX_TAM ]                 ; /* vetor auxiliar */

/* copia matriz A e o vetor B para os auxiliares Aux e Bux para no alterar A e
   B */
   for ( i = 0 ; i < iN ; i++ ) {
      for ( j = 0 ; j < iN ; j++ ) {
         Aux[ i ][ j ] = A[ i ][ j ] ;
      } /* for */
      Bux[ i ] = B[ i ] ;
   } /* for */

/* lao para clculo de L e U */
   for ( i = 0 ; i < iN ; i++ ) {

      for ( j = 0 ; j < iN ; j++ ) {

         if ( i > j ) {  /* se i > j, L[ i ][ j ] != 0 e U[ i ][ j ] = 0  */

         /* calcula L[ i ][ j ] */
            LU[ i ][ j ] = Aux[ i ][ j ] ;
            for ( k = 0 ; k < j ; k++ ) {
               LU[ i ][ j ] -= LU[ i ][ k ] * LU[ k ][ j ] ;
            } /* for */

         /* no h como U[ j ][ j ] == 0, pois se no final de uma iterao de i,
            U[ i ][ i ] == 0, faz-se permutao das linhas de A e repete-se a
            iterao */
            LU[ i ][ j ] /= LU[ j ][ j ] ;

         /* a parte inferior de U  zero */

         } else {  /* se i < j, U[ i ][ j ] != 0 e L[ i ][ j ] = 0  */

         /* calcula U[ i ][ j ] */
            LU[ i ][ j ] = Aux[ i ][ j ] ;
            for ( k = 0 ; k < i ; k++ ) {
               LU[ i ][ j ] -= LU[ i ][ k ] * LU[ k ][ j ] ;
            } /* for */

         /* a parte superior de L  zero */

         } /* if */

      } /* for */

      /* a diagonal de L  1 */

   /* testa se U[ i ][ i ] == 0 para checar a necessidade de permutar as linhas
      de A */
      if ( LU[ i ][ i ] == 0 ) {
         int k        ;   /* contador para permutar linhas */
         double dTemp ;   /* valor temporrio durante a troca */

      /* se o teste  verdadeiro, no h como permutar linhas e A  singular,
         AX = B no tem soluo ou tem vrias solues */
         if ( iProximo == iN ) {
            return 0 ;
         } /* if */

      /* permutao de linhas de A*/
         for ( k = 0 ; k < iN ; k++ ) {
            dTemp = Aux[ i ][ k ] ;
            Aux[ i ][ k ] = Aux[ iProximo ][ k ] ;
            Aux[ iProximo ][ k ] = dTemp ;
         } /* for */

      /* permutao no vetor B */
         dTemp = Bux[ i ] ;
         Bux[ i ] = Bux[ iProximo ] ;
         Bux[ iProximo ] = Bux[ i ] ;

      /* se tiver que permutar uma outra vez, vai ser com as lihnas i e i + 2 */
         iProximo++ ;

         i-- ;  /* vai repetir a iterao com A permutada */
      } else {

      /* se tiver que permutar na prxima iterao, vai ser entre as linhas
         i + 1 e i + 2 */
         iProximo = i + 2 ;

      } /* if */

   } /* for */

   for ( i = 0 ; i < iN ; i++ ) {  /* clculo de LY = B */
      X[ i ] = Bux[ i ] ;

      for ( j = 0 ; j < i ; j++ ) {
         X[ i ] -= X[ j ] * LU[ i ][ j ] ;
      } /* for */

   } /* for */

  for ( i = iN ; i >= 0 ; i-- ) {    /* clculo de Ux = Y */
      X[ i ] = X[ i ] ;

      for ( j = i + 1 ; j < iN ; j++ ) {
         X[ i ] -= X[ j ] * LU[ i ][ j ] ;
      } /* for */

      X[ i ] /= LU[ i ][ i ] ; /* se chegou at aqui, U[ i ][ i ] != 0 */
   } /* for */

   return 1 ;

} /* iLU */


/*==============================================================================
*Nome: ImprimirL
*=============================================================================*/

void ImprimirL( double LU[ ][ I_MAX_TAM ] ,
                int    iN                  ) {

   int i ,  /* contador para linhas */
       j ;  /* contador para colunas */

/* imprime elementos de L */
   for ( i = 0 ; i < iN ; i++ ) {
      for ( j = 0 ; j < iN ; j++ ) {

         if ( i < j ) {
            printf( "%26.22lf", 0. ) ;   /* parte superior de L  0 */
         } else {
            if ( i > j ) {             /* parte inferior de L  != 0 */
               printf( "%26.22lf", LU[ i ][ j ] ) ;
            } else {
               printf( "%26.22lf", 1. ) ; /* diagonal de L  1 */
            } /* if */
         } /* if */

      } /* for */
      printf( "\n" ) ;
   } /* for */

} /* ImprimirL */


/*==============================================================================
*Nome: ImprimirU
*=============================================================================*/

void ImprimirU( double LU[ ][ I_MAX_TAM ] ,
                int    iN                  ) {

   int i ,  /* contador para linhas */
       j ;  /* contador para colunas */

/* imprime elementos de L */
   for ( i = 0 ; i < iN ; i++ ) {
      for ( j = 0 ; j < iN ; j++ ) {

         if ( i > j ) {
            printf( "%26.22lf", 0. ) ;   /* parte inferior de U  0 */
         } else {
            printf( "%26.22lf", LU[ i ][ j ] ) ; /* parte no inferior != 0 */
         } /* if */

      } /* for */
      printf( "\n" ) ;
   } /* for */

} /* ImprimirU */


/*==============================================================================
*Nome: ImprimirVetor
*=============================================================================*/

void ImprimirVetor( double V[ ] ,
                    int    iN    ) {

   int i ;  /* contador */

   printf( "( " ) ;

/* imprime elementos de V */
   for ( i = 0 ; i < iN ; i++ ) {
      printf( "%.17lf ", V[ i ] ) ;
   } /* for */

   printf( " )" ) ;
   printf( "\n" ) ;

} /* ImprimirVetor */


int main( void ) {

/* matriz A dada no teste de AX = B */
   double A[ I_MAX_TAM ][ I_MAX_TAM ]
= { { 1.00, 0.00, 0.00, 0.14, 0.11, 0.09, 0.00, 0.00, 0.00, 0.00, 1.00, 0.00 },
    { 0.00, 1.00, 0.00, 0.00, 0.00, 0.00, 0.33, 0.20, 0.14, 0.00, 0.00, 0.00 },
    { 0.00, 0.00, 1.00, 0.00, 0.00, 0.00, 0.20, 0.14, 0.11, 0.00, 0.00, 0.00 },
    { 0.00, 0.00, 0.00, 1.00, 0.00, 0.00, 0.14, 0.11, 0.09, 0.00, 0.00, 0.00 },
    { 0.00, 0.00, 0.00, 0.00, 1.00, 0.00, 0.00, 0.00, 0.00, 0.33, 0.20, 0.14 },
    { 0.00, 0.00, 0.00, 0.00, 0.00, 1.00, 1.00, 0.00, 0.00, 0.20, 0.14, 0.11 },
    { 0.00, 0.00, 0.00, 0.00, 1.00, 1.00, 1.00, 0.00, 0.00, 0.53, 0.34, 0.26 },
    { 0.14, 0.11, 0.09, 0.00, 0.00, 0.00, 0.00, 1.00, 0.00, 0.00, 0.00, 0.00 },
    { 0.00, 0.00, 0.00, 0.33, 0.20, 0.14, 0.00, 0.00, 1.00, 0.00, 0.00, 0.00 },
    { 0.00, 0.00, 0.00, 0.20, 0.14, 0.11, 0.00, 0.00, 0.00, 1.00, 0.00, 0.00 },
    { 0.20, 0.14, 0.11, 0.00, 0.00, 1.00, 0.00, 0.00, 0.00, 0.00, 1.00, 0.00 },
    { 0.33, 0.20, 0.14, 0.00, 0.00, 0.00, 1.00, 0.00, 0.00, 0.00, 0.00, 1.00 } } ,

/* vetor B de AX = B */
          B[ I_MAX_TAM ]  = { 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1. } ,

/* vetor soluo X de AX = B por LU */
          XLU[ I_MAX_TAM ]                                                     ,

/* matriz L\U */
          LU[ I_MAX_TAM ][ I_MAX_TAM ]                                         ;

/* resultado da possibilidade do clculo da soluo por LU */
   int    iResultado                                                           ;

/* LU */
   iResultado = iLU( A, B, XLU, LU, I_MAX_TAM ) ; /* resolve o sistema por LU */
   if ( iResultado != 0 ) {
      printf( "LU\n\nL:\n" ) ;
      ImprimirL( LU, I_MAX_TAM ) ;         /* imprime a matriz L */
      printf( "\nU:\n" ) ;
      ImprimirU( LU, I_MAX_TAM ) ;         /* imprime a matriz U */
      printf( "\nSoluo: " ) ;
      ImprimirVetor( XLU, I_MAX_TAM ) ;    /* imprime a soluo */
   } else {
      printf( SZ_NAO_UNI ) ;
   } /* if */

   return 0 ;

} /* main */


/* T1AMNPAL.C */
