/*==============================================================================
*NOME:   T1BM3MNPAL.C
*
*  Projeto: Questo B do Teste 1 de Mtodos Numricos Para lgebra Linear
*
*  Verso:              V 1.00
*  Autor :   Juliana Carpes Imperial        - 9714169-4
*
*Descrio do mdulo de implementao
*     Objetivos:
*        - Implementar uma funo para calcular a soluo de um sistema linear
*          por um mtodo iterativo que converge garantidamente para a matriz
*          dada;
*        - Implementar um programa principal que, dado um sistema, aplica
*          as funo acima citada e imprime resultados e comparaes.
*=============================================================================*/

#include <stdio.h>
#include <math.h>

#define I_MAX_TAM 12           /* tamanho mximo da matriz */
#define I_MAX_ITE 1000         /* nmero mximo de iteraes */

/* strings para formatao de sada */
#define SZ_SIDEL  "GAUSS-SIDEL:\n\n"
#define SZ_ERRO   "\nErro: %le\n"
#define SZ_ITE    "Numero de iteracoes: %d"
#define SZ_MAT_R  "R:\n"
#define SZ_VET_K  "\nK: "
#define SZ_RESULT "\nResultado: "

/* definio das funes */

/*==============================================================================
*Nome: iGaussSidel
*Descrio da Funo: funo para calcular a soluo de um sistema linear pelo
*                     mtodo iterativo de Gauss-Sidel.
*
*Parametros recebidos
*   A:  A de AX = B;
*   B:  B de AX = B;
*   X:  X0, valor inicial;
*   iN: o tamanho do sistema.
*
*Parmetros retornados
*   X:     a soluo do sistema se o algoritmo converge;
*   R:     a matriz de Gauss-Sidel;
*   K:     a matriz constante do mtodo;
*   dErro: o erro.
*
*Valor retornado
*   o nmero de iteraes necessrio ou zero, se no convergiu.
*=============================================================================*/

int iGaussSidel( double  A[ ][ I_MAX_TAM ] ,
                 double  B[ ]              ,
                 double  X[ ]              ,
                 double  R[ ][ I_MAX_TAM ] ,
                 double  K[ ]              ,
                 double* dErro             ,
                 int     iN                 ) ;


/*==============================================================================
*Nome: ImprimirR
*Descrio da Funo: funo para imprimir a matriz R de Gauss-Sidel.
*
*Parametros recebidos
*   R:  a matriz de Gauss-Sidel;
*   iN: o tamannho da matriz R.
*=============================================================================*/

void ImprimirR( double R[ ][ I_MAX_TAM ] ,
                int    iN                 ) ;


/*==============================================================================
*Nome: ImprimirVetor
*Descrio da Funo: funo para imprimir um vetor.
*
*Parametros recebidos
*   V:  o vetor;
*   iN: o tamannho do vetor.
*=============================================================================*/

void ImprimirVetor( double V[ ] ,
                    int    iN    ) ;


/*==============================================================================
*Nome: CalcularErro
*Descrio da Funo: funo para calcular o erro entre dois vetores.
*
*Parametros recebidos
*   V:  o vetor;
*   U:  o segundo vetor;
*   iN: o tamannho dos vetores.
*
*Valor retornado
*   o erro entre eles usando norma linha.
*=============================================================================*/

double CalcularErro( double V[ ] ,
                     double U[ ] ,
                     int    iN    ) ;


/*==============================================================================
*Nome: CalcularNorma
*Descrio da Funo: funo para calcular a norma de uma matriz quadrada.
*
*Parametros recebidos
*   A: a matriz;
*   iN: o tamannho da matriz.
*
*Valor retornado
*   a norma linha da matriz.
*=============================================================================*/

double CalcularNorma( double A[ ][ I_MAX_TAM ] ,
                      int    iN                 ) ;


/* implementao das funes */

/*==============================================================================
*Nome: iGaussSidel
*=============================================================================*/

int iGaussSidel( double  A[ ][ I_MAX_TAM ] ,
                 double  B[ ]              ,
                 double  X[ ]              ,
                 double  R[ ][ I_MAX_TAM ] ,
                 double  K[ ]              ,
                 double* dErro             ,
                 int     iN                 ) {

   double X1[ I_MAX_TAM ] ,  /* soluo atual */
          dNorma          ,  /* norma linha da matriz de Gauss-Sidel */
          dCoefMult       ;  /* coeficiente para clculo do erro */
   int    i               ,  /* contador para linhas */
          j               ,  /* contador para colunas */
          k               ;  /* contador de iteraes do Gauss-Sidel */

/* inicializa R e K */
   for ( i = 0 ; i < iN ; i++ ) {
      for ( j = 0 ; j < iN ; j++ ) {
         R[ i ][ j ] = 0. ;
      } /* for */
      K[ i ] = 0. ;
   } /* for */

/* calcula as matrizes R e K de X = JX + K atravs de AX = B */
   for ( i = 0 ; i < iN ; i++ ) {

      if ( A[ i ][ i ] == 0. ) {  /* testa se  h elemento nulo na diagonal */
         return 0 ;
      } /* if */

      for ( j = 0 ; j < iN ; j++ ) {

         if ( i > j ) {
            for ( k = 0 ; k < iN ; k++ ) {
               R[ i ][ k ] -= R[ j ][ k ] * A[ i ][ j ] / A[ i ][ i ] ;
            } /* for */
            K[ i ] -= K[ j ] * A[ i ][ j ] / A[ i ][ i ] ;

         } else {
            if ( i != j ) {
               R[ i ][ j ] -= A[ i ][ j ] / A[ i ][ i ] ;
            } /* if */

         } /* if */

      } /* for */

      K[ i ] += B[ i ] / A[ i ][ i ] ;
   } /* for */

   k = 0 ;                                /* inicializa nmero de iteraes */
   *dErro = 1. ;                          /* inicializa erro */
   dNorma = CalcularNorma( R, iN ) ;      /* calcula norma linha de R */

/* coeficiente multiplicador  dNorma / ( 1 - dNorma ) se dNorma != 1 e igual a
   1 caso contrrio */
   if ( dNorma != 1. ) {
      dCoefMult = dNorma / ( 1. - dNorma ) ;
   } else {
      dCoefMult = 1. ;
   } /* if */

/* continua at erro < 1.e-8 ou se no convergir */
   while ( k < I_MAX_ITE
        && *dErro >= 1.e-8 ) {

   /* XN+1 = R * XN + K */
      for ( i = 0 ; i < iN ; i++ ) {
         X1[ i ] = K[ i ] ;
         for ( j = 0 ; j < iN ; j++ ) {
            X1[ i ] += X[ j ] * R[ i ][ j ] ;
         } /* for */
      } /* for */

   /* calcula erro entre XN+1 e XN */
      *dErro = fabs( dCoefMult * CalcularErro( X, X1, iN ) ) ;

   /* X = X1 */
      for ( i = 0 ; i < iN ; i++ ) {
         X[ i ] = X1[ i ] ;
      } /* for */

      k++ ;      /* incrementa iterao */
   } /* while */

/* se erro <= 1.e-8, convergiu e retorna nmero de iteraes; seno retorna 0 */
   if ( *dErro <= 1.e-8 ) {
      return k ;
   } else {
      return 0 ;
   } /* if */

} /* iGaussSidel */


/*==============================================================================
*Nome: ImprimirR
*=============================================================================*/

void ImprimirR( double R[ ][ I_MAX_TAM ] ,
                int    iN                 ) {

   int i ,  /* contador para linhas */
       j ;  /* contador para colunas */

/* imprime elementos de R */
   for ( i = 0 ; i < iN ; i++ ) {
      for ( j = 0 ; j < iN ; j++ ) {
         printf( "%25.21lf", R[ i ][ j ] ) ;
      } /* for */
      printf( "\n" ) ;
   } /* for */

} /* ImprimirR */


/*==============================================================================
*Nome: ImprimirVetor
*=============================================================================*/

void ImprimirVetor( double V[ ] ,
                    int    iN    ) {

   int i ;  /* contador */

   printf( "( " ) ;

/* imprime elementos de V */
   for ( i = 0 ; i < iN ; i++ ) {
      printf( "%22.18lf", V[ i ] ) ;
   } /* for */

   printf( " )" ) ;
   printf( "\n" ) ;

} /* ImprimirVetor */


/*==============================================================================
*Nome: CalcularErro
*=============================================================================*/

double CalcularErro( double V[ ] ,
                     double U[ ] ,
                     int    iN    ) {

   double dDifer[ I_MAX_TAM ]     ,   /* vetor diferena entre U e V */
          dErro              = 0. ;   /* erro calculado com a norma linha */
   int    i                       ;   /* contador */

/* calcula max| V[ i ] - U[ i ] |, para i entre 0 e iN */
   for ( i = 0 ; i < iN ; i++ ) {
      dDifer[ i ] = fabs( V[ i ] - U[ i ] ) ;
      dErro = dDifer[ i ] > dErro ? dDifer[ i ] : dErro ;
   } /* for */

   return dErro ;  /* retorna o erro  */

} /* CalcularErro */


/*==============================================================================
*Nome: CalcularNorma
*=============================================================================*/

double CalcularNorma( double A[ ][ I_MAX_TAM ] ,
                      int    iN                 ) {

   double dNorma            = 0. ,  /* norma linha de A */
          dSoma[ I_MAX_TAM ]     ;  /* vetor que contm soma das colunas de A */
   int    i                      ,  /* contador para linhas */
          j                      ;  /* contador para colunas */

/* clculo da norma linha */
   for ( i = 0 ; i < iN ; i++ ) {
      for ( j = 0 ; j < iN ; j++ ) {

         if ( j == 0 ) {

         /* inicializa a soma das colunas quando ainda se est no primeiro
            termo */
            dSoma[ i ] = fabs( A[ i ][ j ] ) ;

         } else {
            dSoma[ i ] += fabs( A[ i ][ j ] ) ;
         } /* if */
      } /* for */

   /* acha a maior soma de colunas */
      dNorma = dSoma[ i ] > dNorma ? dSoma[ i ] : dNorma ;

   } /* for */

   return dNorma ;

} /* CalcularNorma */


int main( void ) {

/* matriz A dada no teste de AX = B */
   double A[ I_MAX_TAM ][ I_MAX_TAM ]
= { { 0.14, 0.11, 0.01, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.000 },
    { 0.09, 0.14, 0.01, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.000 },
    { 0.05, 0.02, 0.14, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.01, 0.00, 0.000 },
    { 0.00, 0.00, 0.00, 0.14, 0.11, 0.01, 0.00, 0.00, 0.00, 0.00, 0.00, 0.000 },
    { 0.00, 0.00, 0.01, 0.09, 0.14, 0.01, 0.00, 0.00, 0.00, 0.00, 0.00, 0.000 },
    { 0.00, 0.00, 0.00, 0.05, 0.02, 0.14, 0.00, 0.00, 0.00, 0.00, 0.00, 0.000 },
    { 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.14, 0.11, 0.01, 0.00, 0.00, 0.000 },
    { 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.09, 0.14, 0.01, 0.00, 0.00, 0.000 },
    { 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.05, 0.02, 0.14, 0.00, 0.00, 0.020 },
    { 0.00, 0.00, 0.00, 0.01, 0.00, 0.00, 0.00, 0.00, 0.00, 0.14, 0.11, 0.010 },
    { 0.00, 0.00, 0.00, 0.01, 0.00, 0.00, 0.00, 0.00, 0.00, 0.23, 0.25, 0.005 },
    { 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.09, 0.14, 0.010 } } ,

/* vetor B de AX = B */
          B[ I_MAX_TAM ] = { 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1. } ,

/* vetor soluo inicial X de AX = B por Gauss-Sidel */
          XSidel[ I_MAX_TAM ]
                         = { 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0. } ,

/* matriz de Gauss-Sidel */
          R[ I_MAX_TAM ][ I_MAX_TAM ]                                         ,

/* vetor constante da soluo por Gauss-Sidel */
          K[ I_MAX_TAM ]                                                      ;

/* contadores para imprimir matrizes */
   int    iResultado                                                          ;

/* erro do Gauss-Sidel */
   double dErro                                                               ;

   printf( SZ_SIDEL ) ;

/* Calcula o resultado do sistema pelo mtodo iterativo de Gauss-Sidel */
   iResultado = iGaussSidel( A, B, XSidel, R, K, &dErro, I_MAX_TAM ) ;

/* imprime matriz R */
   printf( SZ_MAT_R ) ;
   ImprimirR( R, I_MAX_TAM ) ;

/* imprime vetor K */
   printf( SZ_VET_K ) ;
   ImprimirVetor( K, I_MAX_TAM ) ;

/* s imprime se convergiu */
   if ( iResultado > 0 ) {

   /* imprime resultado */
      printf( SZ_RESULT ) ;
      ImprimirVetor( XSidel, I_MAX_TAM ) ;

      printf( SZ_ERRO, dErro ) ;        /* imprime erro */
      printf( SZ_ITE, iResultado ) ;    /* imprime nmero de iteraes */

   } /* if */

   return 0 ;

} /* main */

/* T1BM3MNPAL.C */
